import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import java.util.Vector;

/**
 * Ristinolla / Ruudukko

 * Pelin hallinnointi; piirtminen, hiiren ja nppimistn kuuntelu,
 * rastittaminen ja pallottaminen, voittotilanteen tarkistaminen;
 * yleenskin _kaikki_ on dumpattu tnne.
 *
 * @author Samuli Kaipiainen
 * @version 1.0
 */

public class Ruudukko extends JPanel implements Runnable
{
	// yhden ruudun leveys ja korkeus pikselein, sek alkuperinen arvo
	private int ruutu, ruudunkoko;

	// ruudukon leveys ja korkeus ruutuina (eli ei pikselein)
	private int xruutuja, yruutuja;

	// tll hetkell valittu ruutu
	private int xruutu, yruutu;

	// edellinen (viimeksi tehty) siirto
	private int xedellinen, yedellinen;

	// edellinen hiiren/nppimistn yrittm siirto
	private int xklik, yklik;

	// ruudukon sijaintipoikkeama pikselialueen vasemmasta ylkulmasta
	private int xoffset, yoffset;

	// itse ruudukko; 0 tyhj, 1 rasti, 2 pallo
	private byte[][] ruudukko;

	// tmnhetkiset pelaajat (0 dummy, 1 rasti ja 2 pallo)
	private Pelaaja[] pelaaja;

	// pelaajien mahdolliset omat menut (0 dummy, 1 rasti ja 2 pallo)
	private JMenu[] pelaajamenu;

	// pelaajien sijaintihakemisto ja pakkaus (huom piste pakkauksen lopussa, paitsi tyhjn)
	// (alunperin piti laittaa pelaajat omaan pakkaukseen ja hakemistoon, mutta siit tuli
	// ongelmaa koska eivt lytneet tarvitsemiaan Pelaaja- ja Ruudukko-luokkia, ph)
	private String pelaajahakemisto = ".";
	private String pelaajapakkaus = "";

	// onko vuoro rasteilla 1 vai palloilla 2
	private byte vuoro;

	// kuinka monta ruutuskaa tytyy saada perkkin voittaakseen
	private int perkkin;

	// voittiko joku; 0 ei, 1 rasti, 2 pallura; kertoo kuka on ensimmisen kerran voittanut
	private int voitto;

	// uh, voiko pelaamista jatkaa voittamisen jlkeen
	private boolean jatkovoitto;

	// jos voitti, arvottu voittoteksti
	private String voittoteksti;

	// mahdolliset voittosuorat talteen piirtmist varten; kuvaus ks. tarkistaVoitto()
	private int[][] voittorivi;

	// turhat, kyseenalaiset ja tiell olevat apuviivat
	private boolean apuviivat;
	private int[][] apuviiva;

	/**
	 * Lski hiirikuuntelija, koska kuuntelijat tytyy list erikseen hiiren napeille,
	 * vetmiselle ja rullan pyritykselle; nin saadaan yksi yhteinen kuuntelijaluokka
	 */
	private class MyMouseListener extends MouseInputAdapter implements MouseWheelListener
	{
		// hiiren edellinen sijainti vedettess, viimeksi painettu hiiren nappi
		private int x0, y0, nappi;

		/**
		 * Laittaa talteen viimeksi painetun hiiren napin, koska mouseDragged ei itse saa sit selville,
		 * sek painokohdan, jotta mouseDragged tiet mist vetminen alkoi;
		 * plottaa joko rastin tai pallon, jos painettiin vasenta hirulanapulaa
		 *
		 * @param e MouseEvent e
		 */
		public void mousePressed(MouseEvent e)
		{
			nappi = e.getButton();
			x0 = e.getX();
			y0 = e.getY();

			// vain ykksnapilla plotataan
			if (nappi != MouseEvent.BUTTON1) return;

			// estmme negatiivisen jakolaskun nollaan pyristymisen.. uh
			if (x0 < xoffset || y0 < yoffset) return;

			// lasketaan siirto ruutuina
			int x = (x0 - xoffset) / ruutu;
			int y = (y0 - yoffset) / ruutu;

			// estmme saman tien ruudukon ylittvn klikkailun
			// (vaikkakin rastipallo() sen huomaisi eik kelpuuttaisi)
			if (x >= xruutuja || y >= yruutuja) return;

			// rekisterimm yritetyn siirron
			klikSiirto(x, y);
		}

		/**
		 * Vaihtaa valitun ruudun
		 *
		 * @param e MouseEvent e
		 */
		public void mouseMoved(MouseEvent e)
		{
			xruutu = (e.getX() - xoffset) / ruutu;
			yruutu = (e.getY() - yoffset) / ruutu;

			// pysytn ruudukon sispuolella
			if (xruutu < 0) xruutu = 0;
			if (xruutu >= xruutuja) xruutu = xruutuja - 1;
			if (yruutu < 0) yruutu = 0;
			if (yruutu >= yruutuja) yruutu = yruutuja - 1;

			// piirrelln mahdollisia apuviivoja
			apuviivat();

			repaint();
		}

		/**
		 * Scrollaa ruudukkoa sikli kuin oikea hiiren nappi on pohjassa
		 *
		 * @param e MouseEvent e
		 */
		public void mouseDragged(MouseEvent e)
		{
			int x = e.getX();
			int y = e.getY();

			if (nappi == MouseEvent.BUTTON3) {
				xoffset += x - x0;
				yoffset += y - y0;
			}

			// laitetaan hiiren nykyinen sijainti talteen
			x0 = x; y0 = y;

			// pivitetn lopuksi ruudukko
			repaint();
		}

		/**
		 * Zoomaa ruudukkoa hiiren sijainti keskipisteen
		 *
		 * @param e MouseEvent e
		 */
		public void mouseWheelMoved(MouseWheelEvent e)
		{
			int x = e.getX();
			int y = e.getY();
			int zoom = e.getWheelRotation();

			// muutetaan yhden ruudun kokoa
			ruutu -= zoom;

			// jos meni huonoksi, korjataan
			if (ruutu < 1) ruutu = 1;

			// liikutetaan ruudukkoa jotenkin pin suunnilleen
			xoffset = x - (x - xoffset) * ruutu / (ruutu + zoom);
			yoffset = y - (y - yoffset) * ruutu / (ruutu + zoom);

			repaint();
		}
	}

	/**
	 * Nppinksittelij nuolinppinpelaamista varten;
	 * laitetaan tllekin oma privaluokka kun kerran hiirikuuntelijallakin on
	 */
	private class MyKeyListener extends KeyAdapter
	{
		public void keyPressed(KeyEvent e)
		{
			switch (e.getKeyCode()) {
				// ruudukolla surffailu
				case KeyEvent.VK_LEFT:
					xruutu--;
					if (xruutu < 0) xruutu += xruutuja;
					break;
				case KeyEvent.VK_RIGHT:
					xruutu++;
					if (xruutu >= xruutuja) xruutu -= xruutuja;
					break;
				case KeyEvent.VK_UP:
					yruutu--;
					if (yruutu < 0) yruutu += yruutuja;
					break;
				case KeyEvent.VK_DOWN:
					yruutu++;
					if (yruutu >= yruutuja) yruutu -= yruutuja;
					break;
/*
				// ruudukon sijainnin & zoomauksen nollaus (menu hoitaa)
				case KeyEvent.VK_BACK_SPACE:
					ruutu = ruudunkoko;
					xoffset = yoffset = 0;
					break;
*/
				// siirtoyrityksen rekisterinti
				case KeyEvent.VK_ENTER: case KeyEvent.VK_SPACE:
					klikSiirto(xruutu, yruutu);
					break;
			}

			// piirrelln tllkin mahdollisia apuviivoja
			apuviivat();

			// pivitetn ruudukko, jos vaikka on tarvetta
			repaint();
		}
	}

	/**
	 * Luo uuden ruudukon
	 *
	 * @param leveys leveys ruutuina
	 * @param korkeus korkeus ruutuina
	 * @param ruudunkoko yhden ruudun leveys ja korkeus pikselein
	 */
	public Ruudukko(int leveys, int korkeus, int ruudunkoko)
	{
		// leveys, korkeus ja ruudun koko talteen
		this.xruutuja = leveys;
		this.yruutuja = korkeus;
		this.ruutu = this.ruudunkoko = ruudunkoko;

		// luodaan tyhj ruudukkotaulukko
		this.ruudukko = new byte[xruutuja][yruutuja];

		// odotellaan pelaajia ([0] on dummy)
		this.pelaaja = new Pelaaja[3];

		// pelaajien menut tulevat tnne jos niin tahtovat ([0] taas dummy, yhtenisyyden vuoksi)
		this.pelaajamenu = new JMenu[3];

		// asetetaan oletusaloituspelaajat (vaihdetaan menusta jos tahdotaan)
		this.asetaPelaaja(1, new Ihminen());
		this.asetaPelaaja(2, new Tietokone());

		// valittu ruutu keskiruutu
		this.xruutu = (xruutuja - 1) / 2;
		this.yruutu = (yruutuja - 1) / 2;

		// alussa ruudukko on paikoillaan vasemmassa ylkulmassa
		this.xoffset = this.yoffset = 0;

		// viisi perkkin voittaakseen saatava on
		this.perkkin = 5;

		// voitto on voitto on voitto
		this.jatkovoitto = false;

		// "ne on vain tiell" 1 ni, "n on kivat" 2 nt
		this.apuviivat = true;

		// valkoinen kiva tausta
		this.setBackground(Color.WHITE);

		// vhemmn tiell oleva kursori
		// this.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
		// this.setCursor(this.getToolkit().createCustomCursor(this.getToolkit().createImage(new byte[0]), new Point(0, 0), "tyhj"));

		// pyydetn tarpeeksi tilaa koko ruudukolle
		this.setPreferredSize(new Dimension(xruutuja * ruutu, yruutuja * ruutu));

		// tarvitaan jotta nppinksittelij kutsutaan
		this.setFocusable(true);

		// tapahtumaksittelijt hiiren napeille, liikkeelle ja rullalle
		MyMouseListener myMouseListener = new MyMouseListener();
		this.addMouseListener(myMouseListener);
		this.addMouseMotionListener(myMouseListener);
		this.addMouseWheelListener(myMouseListener);

		// tapahtumaksittelij nppimistlle
		MyKeyListener myKeyListener = new MyKeyListener();
		this.addKeyListener(myKeyListener);
	}

	/**
	 * Antaa kopion nykyisen pelitilanteen ruudukosta (kopio jotta kukaan ei srki sit),
	 * miss kopion pyytjn merkit ovat 1 ja vastustajan merkit 2 (tyhjt 0)
	 *
	 * @param pelaaja kopion pyytj
	 * @return ruudukon kopio byte[leveys][korkeus] -taulukkona, miss 0==tyhj, 1==pelaaja, 2==vastustaja
	 */
	public byte[][] annaRuudukonKopio(Pelaaja pelaaja)
	{
		boolean vaihda = false;
		if (pelaaja == this.pelaaja[2]) vaihda = true;

		byte[][] tempruudukko = new byte[xruutuja][yruutuja];

		for (int x = 0; x < xruutuja; x++)
			for (int y = 0; y < yruutuja; y++)
				if (ruudukko[x][y] == 1) tempruudukko[x][y] = (byte) (vaihda ? 2 : 1);
				else if (ruudukko[x][y] == 2) tempruudukko[x][y] = (byte) (vaihda ? 1 : 2);

		return tempruudukko;
	}

	/**
	 * Kertoo onko apuviivat pll
	 *
	 * @return true jos pll, false muuten
	 */
	public boolean onkoApuviivat()
	{
		return this.apuviivat;
	}

	/**
	 * Vaihtaa apuviivat plle ja pois
	 */
	public void vaihdaApuviivat()
	{
		this.apuviivat = !this.apuviivat;
		if (!this.apuviivat) apuviiva = null;
	}

	/**
	 * Kertoo onko jatkovoitto pll
	 *
	 * @return true jos pll, false muuten
	 */
	public boolean onkoJatkovoitto()
	{
		return this.jatkovoitto;
	}

	/**
	 * Vaihtaa jatkovoiton plle ja pois
	 */
	public void vaihdaJatkovoitto()
	{
		this.jatkovoitto = !this.jatkovoitto;
	}

	/**
	 * Alustaa ruudukon sijainnin ja zoomauksen alkuperisiin arvoihin
	 */
	public void nollaaRuudukonSijainti()
	{
		ruutu = ruudunkoko;
		xoffset = yoffset = 0;
		repaint();
	}

	/**
	 * Pelin pyritys -thread; pelaa yhden pelin ja j odottamaan seuraavan alkua, ...
	 */
	public void run()
	{
		while (true) {
			// pelataan yksi peli
			this.peli();

			System.out.println("Ruudukko: peli pttyi");

			// odotetaan uutta
			try { synchronized (this) { wait(); } }
			catch (InterruptedException e) { }
		}
	}

	/**
	 * Pelaa yhden pelin
	 */
	private void peli()
	{
		// kyselln siirtoja niin kauan kuin on jonkun vuoro
		while (vuoro != 0) {
			int[] siirto;

			// Annetaan edellinen siirto ja pyydetn uusi.
			// Lievhk ongelma nousee, jos siirtoa odotellessa vaihdetaan pelaajaa,
			// jonka siirtoa odotellaan; pelaaja vaihtuu vasta seuraavan siirron jlkeen,
			// paitsi jos odoteltiin ihmispelaajan siirtoa, joka palauttaa (-1, -1) -siirron.
			// Lisksi voi tulla ongelmaa, jos aloitetaan uusi peli siirto-kutsun aikana;
			// ihmispelaaja palauttaa ruudukolta saamansa (-1, -1) -siirron, muista ei tied...
			siirto = pelaaja[vuoro].siirto(xedellinen, yedellinen);

			// josko pelaaja luovutti tai bugittaa, ei voida jatkaa peli
			if (siirto == null || siirto.length < 2) {
				System.out.println("Ruudukko: pelaaja " + vuoro + " luovutti, uh");
				break;
			}

			boolean siirtoOK = rastipallo(siirto[0], siirto[1]);

			// debugataan virheellisest siirrosta
			if (!siirtoOK) System.out.println("Ruudukko: pelaaja " + vuoro + " yritti virheellist siirtoa"
				    + " (" + siirto[0] + ", " + siirto[1] + ")");
		}
	}

	/**
	 * Kynnist pelin hallinnointithreadin ja aloittaa uuden pelin
	 */
	public void kynnist()
	{
		// ok, uusi peli vain kyntiin!
		this.uusiPeli();

		// threadi plle, hieman pienempi prioriteetti niin saadaan piirt rauhassa
		Thread peli = new Thread(this);
		peli.setPriority(peli.getPriority() - 1);
		peli.start();
	}

	/**
	 * Tappaa ruudukon ja ristinollan ja kaiken, mennn pois
	 */
	public void lopeta()
	{
		System.exit(0);
	}

	/**
	 * Aloittaa uuden pelin; luo uuden ruudukon ja nollaa pelikohtaiset muuttujat,
	 * ilmoittaa uudesta pelist pelaajille sek pelin hallinnointithreadille
	 */
	public void uusiPeli()
	{
		// nollataan ruudukko
		for (int x = 0; x < ruudukko.length; x++)
			for (int y = 0; y < ruudukko[x].length; y++)
				ruudukko[x][y] = 0;

		// nollataan edellinen siirto
		xedellinen = yedellinen = -1;

		// nollataan mys edellinen hiiri/nppimist -siirtoyritys, jottei tule
		// ylimrist siirtoa uuden pelin alkuun kesken olleen pelin takia
		xklik = yklik = -1;

		// mahdollinen voittotilanne pois plt
		voitto = 0;
		voittorivi = new int[4][];
		voittoteksti = null;

		// alussa vuoro rasteilla
		vuoro = 1;

		// ilmoitetaan pelaajille uudesta pelist sek ruudukon koko
		pelaaja[1].uusiPeli();
		pelaaja[2].uusiPeli();

		// kun viel pivitetn ruudukon piirtotilanne niin hyv
		repaint();

		// ja kerrotaan peli-threadille jotta aloittaa uuden pelin, jos on tauolla;
		// jos on peli kynniss, hertetn mahdollisesti odotteleva annaSiirto(), ja
		// saadaan kontrolli takaisin pelithreadiin
		synchronized (this) { notify(); }
	}

	/**
	 * Etsii kaikki pelaajat-hakemistosta lytyvt pelaajat
	 *
	 * @return lydetyt pelaajat Pelaaja-Classeina, null jos tuli ongelmallinen ongelma
	 */
	public Class[] etsiPelaajat()
	{
		String[] ps;

		// katsotaan saadaanko lukea tiedostolistaus; appletti ei saa eik voisikaan
		try {
			// etsitn kaikki .class-tiedostot pelaajahakemistosta
			ps = new File(pelaajahakemisto).list(new FilenameFilter()
			{
				public boolean accept(File dir, String name)
				{
					if (name.endsWith(".class")) return true;
					else return false;
				}
			});
		} catch (SecurityException e) {
			// ei saatu lukea hakemistoa, keksitn jotain muuta
			ps = new String[] { "Ihminen.class", "Satu.class", "Satuhali.class", "Tietokone.class" };
		}

		// jos tuli ongelma, ei siis lytynyt pelaajiakaan eik voida kyll pelatakaan
		if (ps == null) {
			System.out.println("HUH, ei lytynyt yhtn pelaajaa, ei voida pelata!");
			return null;
		}

		// tnne lydetyt pelaajat talteen (kytmme Vectoria, koska meneilln on
		// tietokonepelaajan toteuttamisen kannalta epoleelliset riennot)
		Vector p = new Vector();

		// kydn lydetyt luokat lpi ja katsotaan, kelpaako pelaajaksi
		for (int n = 0; n < ps.length; n++)
		{
			// poistetaan .class-pte
			ps[n] = ps[n].substring(0, ps[n].lastIndexOf(".class"));

			try {
				// luodaan luokka, tarvitaan mahdollinen pakkaus mukaan
				Class pc = Class.forName(pelaajapakkaus + ps[n]);

				// jos kyseess on ihan oikea Pelaaja-luokka, laitetaan talteen
				if (Class.forName("Pelaaja").isAssignableFrom(pc) && !pc.isInterface())
					p.add(pc);

			} catch (ClassNotFoundException e) {
				System.out.println(e);
			}
		}

		// debugataan lydetyt pelaajat
		System.out.print("Lydetyt pelaajat:");
		for (int n = 0; n < p.size(); n++)
			System.out.print(" " + ((Class) p.get(n)).getName());
		System.out.println();

		// jostain &)%##")# syyst tm palauttaa ClassCastExceptionin
		// return (Class[]) p.toArray();

		// ...joten; harmittaa kun vaihdoin taulukon+laskurin Vectoriksi, vaan en ala en
		Class[] pelaaja = new Class[p.size()];
		for (int n = 0; n < p.size(); n++) pelaaja[n] = (Class) p.get(n);

		return pelaaja;
	}

	/**
	 * Vaihtaa pelaajan (rasti tai pallo) uudeksi, sek luo ja alustaa kyseisen pelaajan
	 *
	 * @param pelaaja vaihdetaanko rasteja==1 vai palloja==2
	 * @param uusipelaaja uusi pelaaja Class-luokkana
	 * @return onnistuiko pelaajan vaihto
	 */
	public boolean asetaPelaaja(int pelaaja, Class uusipelaaja)
	{
		Pelaaja p;

		// yritetn luoda uusi pelaaja
		try {
			p = (Pelaaja) uusipelaaja.newInstance();
		} catch (IllegalAccessException e) {
			System.out.println(e);
			return false;
		} catch (InstantiationException e) {
			System.out.println(e);
			return false;
		}

		// vaihdetaan ja alustetaan
		return asetaPelaaja(pelaaja, p);
	}

	/**
	 * Vaihtaa pelaajan (rasti tai pallo) uudeksi, sek alustaa kyseisen pelaajan
	 *
	 * @param pelaaja vaihdetaanko rasteja==1 vai palloja==2
	 * @param uusipelaaja uusi pelaaja
	 * @return onnistuiko pelaajan vaihto
	 */
	public boolean asetaPelaaja(int pelaaja, Pelaaja uusipelaaja)
	{
		// ei vaihdella olemattomia
		if (pelaaja != 1 && pelaaja != 2) return false;

		// vaihdetaan, nollataan pelaajan menu ja kerrotaan ruudukon tiedot
		// (pelaaja kertoo asetaRuudukko-kutsulla oman menunsa, jos hnell sellainen on)
		this.pelaaja[pelaaja] = uusipelaaja;
		this.pelaajamenu[pelaaja] = null;
		uusipelaaja.asetaRuudukko(this, xruutuja, yruutuja);

		// jos on peli kynniss, hertetn mahdollisesti odotteleva annaSiirto(),
		// jotta saadaan kontrolli takaisin pelithreadiin ja pelaaja vaihtuu heti
		if (vuoro != 0) synchronized (this) { notify(); }

		return true;
	}

	/**
	 * Antaa nykyisen pelaajan, rastin tai pallon, Class-luokkana
	 *
	 * @param pelaaja annetaanko rasti==1 vai pallo==2
	 * @return nykyinen pyydetty pelaaja, null jos pyydettiin hulluja
	 */
	public Class annaPelaaja(int pelaaja)
	{
		if (pelaaja != 1 && pelaaja != 2) return null;

		return this.pelaaja[pelaaja].getClass();
	}

	/**
	 * Asettaa pelaajamenun; pmenun tulee kysell kyseist menua annaPelaajaMenulla
	 * luodessaan itsens ja aina pelaajaa vaihdettaessa, pelaajan tulee kertoa oma
	 * menunsa asetaRuudukko-kutsun aikana, jotta menu ehtii mukaan pmenuun
	 *
	 * @param pelaaja kuka kertoo menunsa olemassaolosta
	 * @param menu kyseisen pelaajan oma menu
	 */
	public void asetaPelaajaMenu(Pelaaja pelaaja, JMenu menu)
	{
		if (pelaaja == this.pelaaja[1]) this.pelaajamenu[1] = menu;
		if (pelaaja == this.pelaaja[2]) this.pelaajamenu[2] = menu;
	}

	/**
	 * Antaa pyydetyn pelaajan oman menun
	 *
	 * @param pelaaja kenen menu halutaan; 1==rasti, 2==pallo
	 * @return pyydetyn pelaajan menu
	 */
	public JMenu annaPelaajaMenu(int pelaaja)
	{
		if (pelaaja != 1 && pelaaja != 2) return null;

		return pelaajamenu[pelaaja];
	}

	/**
	 * Kertoo (ihmispelaajalle) seuraavan hiirelt/nppimistlt luetun ihmisen tekemn siirron
	 *
	 * @return int[2] {x, y}; {-1, -1} jos peli lopetettiin siirtoa odotellessa
	 */
	public int[] annaSiirto()
	{
		// odotellaan jotta saadaan uusi siirtoyritys eteenpin kerrottavaksi
		try {
			synchronized (this) { wait(); }
		} catch (InterruptedException e) { }

		return new int[] {xklik, yklik};
	}

	/**
	 * Laittaa talteen edellisen hiirelt/nppimistlt saadun siirron annaSiirtoa varten
	 *
	 * @param x siirron x-ruutu
	 * @param y siirron y-ruutu
	 */
	private void klikSiirto(int x, int y)
	{
		// siirto talteen
		this.xklik = x;
		this.yklik = y;

		// hertetn mahdollisesti odotteleva annaSiirto()
		synchronized (this) { notify(); }
	}

	/**
	 * Asettaa ruudukkoon rastin tai pallon (sikli kuin siirto on sallittu)
	 *
	 * @param x x-koordinaatti ruutuina
	 * @param y y-koordinaatti ruutuina
	 *
	 * @return true jos kaikki ok, false jos ruudussa ennestn tavaraa tai yritettiin ruudukon ulkopuolelle
	 */
	private boolean rastipallo(int x, int y)
	{
		try {
			if (ruudukko[x][y] != 0) return false;
		} catch (ArrayIndexOutOfBoundsException e) {
			return false;
		}

		// ok, pivitetn ruudukko
		ruudukko[x][y] = vuoro;

		// laitetaan siirto muistiin
		xedellinen = x; yedellinen = y;

		// tarkistetaan voittiko edellinen siirto, ekan kerran tai jatkoajalla;
		// pivitetn voittoteksti, jos joku juuri voitti
		if (voitto == 0 && tarkistaVoitto(x, y)) {
			voitto = vuoro;
			voittoteksti = voittoteksti(vuoro);
		} else if (jatkovoitto) {
			// jatkoajalla nytetn edellist voittosuoraa kunnes tulee uusi voitto
			int[][] temp_voittorivi = voittorivi;
			if (tarkistaVoitto(x, y)) voittoteksti = voittoteksti(vuoro);
			else voittorivi = temp_voittorivi;
		}

		// vaihdetaan vuoro seuraavalle, jollei ole jo voitettu lopullisesti, jolloin peli seis
		if (voitto == 0 || jatkovoitto) vuoro ^= 3;
		else vuoro = 0;

		// tllkin apuviivat
		apuviivat();

		// piirretn ruudukko
		repaint();

		return true;
	}

	/**
	 * Tarkistaa, muodostaako annettu piste ympristineen voittotilanteen, eli
	 * 5 samaa merkki perkkin johkin suuntaan kyseisen pisteen ymprill;
	 * lisksi laittaa voittorivi-taulukkoon talteen mahdollisten voittosuorien
	 * alku- ja loppu-koordinaatit
	 *
	 * @param x tarkasteltavan pisteen x-koordinaatti
	 * @param y ... y-koordinaatti
	 * @return true jos voitto lytyi, false ingen vinst
	 *
	 * voittorivi[4][4] [n][ab*xy]-taulukko, jossa mahdollisten voittosuorien alku- (vasen-yl) ja
	 * loppu (oikea-ala) -koordinaatit; kukin voittorivi[n] null, jos ei kyseist voittosuoraa
	 *
	 * n==0 lnsi-it, n==1 pohjoinen-etel, n==2 luode-kaakko, n==3 lounas-koillinen
	 *
	 * [0][0] lnsi-it -suoran alku-x
	 * [0][1] lnsi-it -suoran alku-y
	 * [0][2] lnsi-it -suoran loppu-x
	 * [0][3] lnsi-it -suoran loppu-y
	 * ...
	 * [1] null, jos ei pohjoinen-etel -voittosuoraa
	 * ...
	 * selv homma?)
	 */
	private boolean tarkistaVoitto(int x, int y)
	{
		// viimeksi pelattu merkki (rasti vai pallo)
		int merkki = ruudukko[x][y];

		// nill laskemme suorien pituuksia
		int a, b;

		// thn talletamme lydetyt voittosuorat; alustuu nulliksi, mist tykkmme
		int[][] vr = new int[4][];

		// lnsi-it (viimeksi pelatusta merkist vasemmalle ja oikealle)
		for (a = x - 1; a >= 0 && ruudukko[a][y] == merkki; a--);
		for (b = x + 1; b < xruutuja && ruudukko[b][y] == merkki; b++);
		if (b - a > perkkin) vr[0] = new int[] {a + 1, y, b - 1, y};

		// pohjoinen-etel
		for (a = y - 1; a >= 0 && ruudukko[x][a] == merkki; a--);
		for (b = y + 1; b < yruutuja && ruudukko[x][b] == merkki; b++);
		if (b - a > perkkin) vr[1] = new int[] {x, a + 1, x, b - 1};

		// luode-kaakko (vhn eri systeemill kuin helpommat vaaka-pysty-suorat; nyt
		// a ja b merkkaavat etisyytt xy-pisteest, eivtk itse x- tai y-pistett)
		for (a = 1; x - a >= 0 && y - a >= 0 && ruudukko[x-a][y-a] == merkki; a++);
		for (b = 1; x + b < xruutuja && y + b < yruutuja && ruudukko[x+b][y+b] == merkki; b++);
		if (a + b > perkkin) vr[2] = new int[] {x-a + 1, y-a + 1, x+b - 1, y+b - 1};

		// lounas-koillinen
		for (a = 1; x - a >= 0 && y + a < yruutuja && ruudukko[x-a][y+a] == merkki; a++);
		for (b = 1; x + b < xruutuja && y - b >= 0 && ruudukko[x+b][y-b] == merkki; b++);
		if (a + b > perkkin) vr[3] = new int[] {x-a + 1, y+a - 1, x+b - 1, y-b + 1};

		// mahdolliset voittosuorat talteen piirtmist varten
		this.voittorivi = vr;

		// katsomma viel siis, lytyik yhtn voittorivi
		for (int n = 0; n < 4; n++) if (vr[n] != null) return true;
		return false;
	}

	/**
	 * Purkka-apuviiva-feature; piirtelee vuorollaan olevan rastin/pallon valitun ruudun
	 * vieress olevien rykelmryhmien plle apuviivoja perkkisten suorien plle,
	 * jolloin nkee mihin on muodostumassa voittosuoria!
	 */
	private void apuviivat()
	{
		// jos apuviivat eivt ole pll, poispoispois
		if (!apuviivat) return;

		// jos ei ole kenenkn vuoro tai valittu ruutu vastustajan merkin pll, ei apuviivoja
		if (vuoro == 0 || ruudukko[xruutu][yruutu] == (vuoro ^ 3)) {
			apuviiva = null;
			return;
		}

		// oikeat arvot talteen
		int temp_perkkin = perkkin;
		byte temp_merkki = ruudukko[xruutu][yruutu];
		int[][] temp_voittorivi = voittorivi;

		// apuviivoja heti jos vain 2 perkkin saadaan
		perkkin = 2;
		ruudukko[xruutu][yruutu] = vuoro;

		// lasketaan apuviivat ja laitetaan talteen
		tarkistaVoitto(xruutu, yruutu);
		apuviiva = voittorivi;

		// palautetaan apuviivamuutetut muuttujat
		voittorivi = temp_voittorivi;
		ruudukko[xruutu][yruutu] = temp_merkki;
		perkkin = temp_perkkin;
	}

	/**
	 * Arpoo ylen hauskan voittotekstin jomman kumman rastin tai pallon voittamisen kunniaksi
	 *
	 * @param voittaja kehutaanko rastia==1 vai palloa==2
	 * @return arvottu teksti
	 */
	private String voittoteksti(int voittaja)
	{
		String rastit[] = {"rasti", "raksuli", "rastuli", "rastila", "raksu", "r",
				  "ruksi", "risti", "raksi"};
		String pallot[] = {"pallo", "pallero", "pallura", "ponkula", "rinkula", "ympyr", "nolla",
				  "renkula", "lonkero", "pullero", "p"};
		String voitot[] = {"voitti", "on voittanut", "meni jatkoon", "on paras", "ihanaa iihaanaaa",
				  "ihanaa", "suoriutui", "teki itsens", "vei rahat", "supermies"};
		String huudot[] = {".", "!", "!!!", "!!111", " (harmi).", "?", " ^^", " O.o",
				  " x)", " ^_^", " o.O"};

		StringBuffer s = new StringBuffer();

		if (voittaja == 1) s.append(rastit[(int) (rastit.length * Math.random())]);
		if (voittaja == 2) s.append(pallot[(int) (pallot.length * Math.random())]);
		s.append(" ");
		s.setCharAt(0, (char) (s.charAt(0) - ' '));

		s.append(voitot[(int) (voitot.length * Math.random())]);
		s.append(huudot[(int) (huudot.length * Math.random())]);

		return s.toString();
	}

	/**
	 * Pivitt ruudukon
	 *
	 * @param g mursu
	 */
	public void paintComponent(Graphics g)
	{
		// annetaan Swingin pyyhki tausta
		super.paintComponent(g);

		// ruudukon rajat
		int x1 = xoffset;
		int y1 = yoffset;
		int x2 = x1 + xruutuja * ruutu;
		int y2 = y1 + yruutuja * ruutu;
/*
		// alpha-kanava (ilkeit synkronointiongelmia jos luetaan samalla kun tietokone pivitt)
		if (((Tietokone) pelaaja[2]).ralpha != null) for (int y = 0; y < yruutuja; y++)
			for (int x = 0; x < xruutuja; x++) {
				int c = 0xff - ((Tietokone) pelaaja[2]).ralpha[x][y] * 0x10;
				if (c < 0) c = 0;
				g.setColor(new Color(c, 0xff, c));
				g.fillRect(x1 + x * ruutu + 1, y1 + y * ruutu + 1, ruutu - 1, ruutu - 1);
			}
*/
		// piirretn ruudukko, jesjes
		g.setColor(new Color(0xaaaaaa));
		for (int y = 0; y <= yruutuja; y++)
			g.drawLine(x1, y1 + y * ruutu, x2, y1 + y * ruutu);
		for (int x = 0; x <= xruutuja; x++)
			g.drawLine(x1 + x * ruutu, y1, x1 + x * ruutu, y2);

		// korostetaan valittu ruutu
		if (vuoro == 1) g.setColor(new Color(0xffddcc));
		else if (vuoro == 2) g.setColor(new Color(0xccddff));
		g.fillRect(x1 + xruutu * ruutu + 1, y1 + yruutu * ruutu + 1, ruutu - 1, ruutu - 1);

		// suunnilleen ruudun kokoinen fontti (pikku-x:n ja o:n osalta)
		g.setFont(new Font("Arial", Font.PLAIN, 2 * ruutu));

		// piirretn rastit ja pallot
		for (int y = 0; y < yruutuja; y++)
			for (int x = 0; x < xruutuja; x++) {
				// lihavoidaan edellinen rasti/pallo
				if (x == xedellinen && y == yedellinen) g.setFont(g.getFont().deriveFont(Font.BOLD));
				if (ruudukko[x][y] == 1) {
					g.setColor(new Color(0xff0000));
					g.drawString("x", x1 + x * ruutu, y1 + (y + 1) * ruutu);
				} else if (ruudukko[x][y] == 2) {
					g.setColor(new Color(0x0000ff));
					g.drawString("o", x1 + x * ruutu, y1 + (y + 1) * ruutu);
				}
				if (x == xedellinen && y == yedellinen) g.setFont(g.getFont().deriveFont(Font.PLAIN));
			}

		// piirretn mahdolliset voittosuorat
		g.setColor(new Color(0x004400));
		if (voittorivi != null) for (int n = 0; n < 4; n++) {
			int[] r = voittorivi[n];
			if (r != null) {
				// piirrelln monta viivaa
				int xa = xoffset + ruutu / 2 + ruutu * r[0];
				int ya = yoffset + ruutu / 2 + ruutu * r[1];
				int xb = xoffset + ruutu / 2 + ruutu * r[2];
				int yb = yoffset + ruutu / 2 + ruutu * r[3];
				g.drawLine(xa, ya, xb, yb);
				g.drawLine(xa, ya - 1, xb, yb - 1);
				g.drawLine(xa + 1, ya, xb + 1, yb);
				g.drawLine(xa + 1, ya - 1, xb + 1, yb - 1);
			}
		}

		// piirretn mahdolliset apuviivat
		if (vuoro == 1) g.setColor(new Color(0x990000));
		if (vuoro == 2) g.setColor(new Color(0x000099));
		if (apuviiva != null) for (int n = 0; n < 4; n++) {
			int[] r = apuviiva[n];
			if (r != null) {
				// piirrelln monta viivaa
				int xa = xoffset + ruutu / 2 + ruutu * r[0];
				int ya = yoffset + ruutu / 2 + ruutu * r[1];
				int xb = xoffset + ruutu / 2 + ruutu * r[2];
				int yb = yoffset + ruutu / 2 + ruutu * r[3];
				g.drawLine(xa, ya, xb, yb);
				g.drawLine(xa, ya - 1, xb, yb - 1);
				g.drawLine(xa + 1, ya, xb + 1, yb);
				g.drawLine(xa + 1, ya - 1, xb + 1, yb - 1);
			}
		}

		// kerrotaan josko joku voitti, ruudukon leveydest riippuvalla fonttikoolla
		// sek hienolla varjoefektill
		if (voittoteksti != null) {
			int w = (xruutuja * ruudunkoko) / 14;
			g.setFont(new Font("Arial", Font.BOLD, w));
			g.setColor(new Color(0x000000));
			g.drawString(voittoteksti, w / 4, w + 2);
			g.setColor(new Color(0x00dd00));
			g.drawString(voittoteksti, w / 4 + 2, w);
		}
/*
		// kerrotaan josko joku voitti, zoomauksesta riippumattomalla fonttikoolla
		// ja hienolla varjoefektill
		if (voittoteksti != null) {
			g.setFont(new Font("Arial", Font.BOLD, ruudunkoko * 4));
			g.setColor(new Color(0x000000));
			g.drawString(voittoteksti, ruudunkoko, 4 * ruudunkoko + 2);
			g.setColor(new Color(0x00dd00));
			g.drawString(voittoteksti, ruudunkoko + 2, 4 * ruudunkoko);
		}
*/
	}
}
